package board;

import java.awt.event.*;
import java.awt.*;
import java.util.*;

public class Lasso
{
	public static Color LASSO_BASE_COLOR = new  Color(255/255f, 255/255f, 200/255f, 0.7f);	
	Stroke stroke;
	DrawPanel drawPanel;
	
	Vector grabbed_strokes;
	
	Lasso(){};
	Lasso(DrawPanel drawPanel, Stroke stroke){
		this.drawPanel = drawPanel;
		this.stroke = stroke;
		
		stroke.points.addElement(new Point((Point)stroke.points.elementAt(0)));
		
		stroke = StrokeAnalyzer.resample_by_length(stroke, 20);
		
		
		grabbed_strokes = new Vector();
		Vector deleted_strokes = new Vector();
		
		Record record = drawPanel.record;
		
		//Vector strokes = record.getStrokes();
		
		//Vector strokes = drawPanel.view.visible_strokes;
		Vector strokes = drawPanel.view.get_visible_strokes();
		for(int i=0; i<strokes.size(); i++){
			Stroke s = (Stroke) strokes.elementAt(i);
			Stroke ss = drawPanel.view.get_screen_stroke(s);
			if (ss.closedBy(stroke)){
				grabbed_strokes.addElement(ss);
				deleted_strokes.addElement(s);	
			}
		}
		
		drawPanel.record.removeStrokes(deleted_strokes);
		
		System.out.println("grabbed "+grabbed_strokes.size());
	}

	
	
	
	
	public void paint(Graphics g){
		
		Polygon polygon = stroke.get_polygon();
		g.setColor(LASSO_BASE_COLOR);
		g.fillPolygon(polygon);
		
		
		
		stroke.paint(g);
		
		for(int i=0; i<grabbed_strokes.size(); i++){
			Stroke s = (Stroke) grabbed_strokes.elementAt(i);
			s.paint(g);
		}
		
		g.setClip(drawPanel.getBounds());
	}
	
	public void drag(int dx, int dy){
		stroke.move_relative(dx, dy);		
		for(int i=0; i<grabbed_strokes.size(); i++)
			((Stroke)grabbed_strokes.elementAt(i)).move_relative(dx, dy);
	}
	
	public void release(){
		//PlaySound.play("click.wav");
		Vector strokes = new Vector();
		for(int i=0; i<grabbed_strokes.size(); i++){
			Stroke ss = (Stroke) grabbed_strokes.elementAt(i);
			drawPanel.view.addStroke(ss);
			//strokes.addElement(s);
		}
		//drawPanel.record.addStrokes(strokes);			
	}
	
	
	
	
	
	
	
	
	public boolean grab_boundary(Point p){
		double d = StrokeAnalyzer.distance(stroke, p);
		return (d<10);
	}
	
	Point center;
	Point start_p;
	Vector original_strokes;
	Stroke original_stroke;
	public void start_resizing(Point p){
		center = stroke.get_center();
		start_p = p;
		
		original_strokes = new Vector();
		for(int i=0; i<grabbed_strokes.size(); i++){
			Stroke s = (Stroke) grabbed_strokes.elementAt(i);
			original_strokes.addElement(s.get_duplicated());
		}
		
		original_stroke = stroke.get_duplicated();
	}
	public void resizing(Point p){
		double scale = Vertex2D.distance(center, p) / Vertex2D.distance(center, start_p);
		double angle = Vector2D.get_angle_360(new Vector2D(center, start_p),  new Vector2D(center, p));
		angle = (int)  (angle/45+0.5)*45;
		grabbed_strokes = new Vector();
		for(int i=0; i<original_strokes.size(); i++){
			Stroke s = (Stroke) original_strokes.elementAt(i);
			grabbed_strokes.addElement(s.get_scaled_and_rotated(center, scale, angle));
			//grabbed_strokes.addElement(s.get_scaled(center, scale));
		}
		
		//stroke = original_stroke.get_scaled(center, scale);
		stroke = original_stroke.get_scaled_and_rotated(center, scale, angle);
		stroke.color = Color.red;
	}
	

	public final static int[] cursors = {Cursor.N_RESIZE_CURSOR,
										 Cursor.NE_RESIZE_CURSOR,
										 Cursor.E_RESIZE_CURSOR,
										 Cursor.SE_RESIZE_CURSOR,
										 Cursor.S_RESIZE_CURSOR,
										 Cursor.SW_RESIZE_CURSOR,
										 Cursor.W_RESIZE_CURSOR,
										 Cursor.NW_RESIZE_CURSOR};
										 
	public Cursor get_cursor(Point p){
		Point center = new Point(stroke.bbox.x+stroke.bbox.width/2, 
								 stroke.bbox.y+stroke.bbox.height/2);
		Vector2D dv = new Vector2D(center, p);
		double angle = Vector2D.get_angle_360(new Vector2D(0,-1),dv);	// clockwise
		angle += 22.5;
		angle = angle - 360*((int)(angle/360));
		double R = 45;
		int n = (int)(angle / R);
		
		return new Cursor(cursors[n]);
	}
	
	
	
}
