package board;

import java.awt.*;
import java.util.*;
import java.io.*;

public class Stroke implements Serializable
{
	public Vector points;
	public int width = 1;
	public Rectangle bbox;
	
	public Stroke(){
		points = new Vector();
		bbox = new Rectangle(-1,-1,-1,-1);
		
		color=Color.black;
	}
	public Stroke(Color c){
		this();
		color = c;
	}
	public Stroke(int x, int y){
		color=Color.black;

		points = new Vector();
		bbox = new Rectangle(-1,-1,-1,-1);
		addPoint(new Point(x,y));		
	}
	public Stroke(Vector points){
		color=Color.black;

		this.points = points;
		update_bbox();
	}
	public Stroke(Point p0, Point p1, Color c){
		this(p0,p1);
		color = c;
	}
	public Stroke(Point p0, Point p1){
		color=Color.black;

		points = new Vector();
		bbox = new Rectangle(-1,-1,-1,-1);
		addPoint(new Point(p0));		
		addPoint(new Point(p1));	
	}
	
	public Stroke get_duplicated(){
		Stroke stroke = new Stroke();
		for(int i=0; i< points.size(); i++){
			Point p = (Point)points.elementAt(i);
			if (p==null)
			stroke.points.addElement(null);
			else
			stroke.points.addElement(new Point(p));
		}
		stroke.bbox = new Rectangle(bbox);
		return stroke;
	}
	
	
	
	
	public int size(){
		return points.size();
	}
	
	
	
	
	public static Stroke move(Stroke stroke, Point v){
		Stroke new_stroke = new Stroke();
		for(int i=0; i< stroke.points.size(); i++){
			Point p = (Point)stroke.points.elementAt(i);
			new_stroke.addPoint(p.x+v.x, p.y+v.y);
		}
		new_stroke.update_bbox();
		return new_stroke;
	}
	
	public static Stroke scale(Stroke stroke, double m){
		Stroke new_stroke = new Stroke();
		for(int i=0; i< stroke.points.size(); i++){
			Point p = (Point)stroke.points.elementAt(i);
			new_stroke.addPoint((int)(p.x*m), (int)(p.y*m));
		}
		new_stroke.update_bbox();
		return new_stroke;
	}
	
	
	public Stroke get_scaled_and_rotated(Point center, double scale, double angle){
		Stroke new_stroke = new Stroke();
		for(int i=0; i< points.size(); i++){
			Point p = (Point)points.elementAt(i);
			Vector2D v = new Vector2D(center, p);
			v.multiply(scale);
			v = Vector2D.rotate(v, angle);
			
			new_stroke.addPoint((int)(center.x+v.x), (int)(center.y+v.y));
		}
		new_stroke.update_bbox();
		return new_stroke;
	}
	public Stroke get_scaled(Point center, double scale){
		Stroke new_stroke = new Stroke();
		for(int i=0; i< points.size(); i++){
			Point p = (Point)points.elementAt(i);
			Vector2D v = new Vector2D(center, p);
			v.multiply(scale);
			//v = Vector2D.rotate(v, angle);
			
			new_stroke.addPoint((int)(center.x+v.x), (int)(center.y+v.y));
		}
		new_stroke.update_bbox();
		return new_stroke;
	}
	
	
	
	
	
	
	
	
	
	public void addEdge(Point p0, Point p1){
		addPoint(new Point(p0));
		addPoint(new Point(p1));
		addPoint(null);
	}
	
	
	public Point firstPoint(){
		return (Point) points.elementAt(0);
	}
	public Point lastPoint(){
		return (Point) points.elementAt(points.size()-1);
	}
	public void addPoint(int x, int y){
		addPoint(new Point(x, y));
	}				   
	public void addPoint(Point p){
		points.addElement(p);
		
		if (p==null)
			return;
		
		if (bbox.x == -1 || bbox.x > p.x-width/2){ 
			int x0 = p.x-width/2;
			bbox.width -= x0-bbox.x;
			bbox.x = x0; 
		}
		if (bbox.y == -1 || bbox.y > p.y-width/2) {
			int y0 = p.y - width/2;
			bbox.height -= y0-bbox.y;
			bbox.y = y0;
		}
		if (bbox.width == -1 || bbox.x + bbox.width < p.x+width/2) 
			bbox.width = p.x+width/2 - bbox.x;
		if (bbox.height == -1 || bbox.y + bbox.height < p.y+width/2) 
			bbox.height = p.y+width/2 - bbox.y;
		
	}

	
	
	
	
	public Rectangle get_bbox(){
		int min_x = -1;
		int min_y = -1;
		int max_x = -1;
		int max_y = -1;
		for(int i=0; i<points.size(); i++){
			Point p = (Point) points.elementAt(i);
			if (min_x == -1 || p.x < min_x)	min_x = p.x;
			if (min_y == -1 || p.y < min_y)	min_y = p.y;
			if (max_x == -1 || p.x > max_x)	max_x = p.x;
			if (max_y == -1 || p.y > max_y)	max_y = p.y;
		}
		return new Rectangle(min_x, min_y, max_x-min_x, max_y-min_y);
	}
	public Point get_center(){
		Rectangle bbox = get_bbox();
		return new Point(bbox.x + bbox.width/2, bbox.y + bbox.height/2);
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	public Color color;//=Color.black;
	BasicStroke basicStroke;
	public void paint(Graphics g){

		//Graphics2D g2 = (Graphics2D)g;
		//g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		//g2.setStroke(new BasicStroke(0.2f));
		//width = 1;

		/*
		Graphics2D g2 = (Graphics2D)g;
		if (basicStroke == null){
			basicStroke =  (new BasicStroke(width, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND ));
		}
		g2.setStroke(basicStroke);
		*/
		
		if (color == null)
			return;
		g.setColor(color);
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if(p0!=null && p1!=null){
				//if (width != 1)
				//	MyGraphics.drawWideLine(g, p0, p1, width);
				//else
					g.drawLine(p0.x, p0.y, p1.x, p1.y);
			}
		}
		if (points.size() == 1){
			Point p0 = (Point)points.elementAt(0);
				//if (width != 1)
				//	MyGraphics.drawWideLine(g, p0, p0, width);
				//else
					g.drawLine(p0.x, p0.y, p0.x, p0.y);
		}
		
		
		//paint_graph(g);
	}
	public void shifted_paint(Graphics g, int x, int y){
		if (color == null)
			return;

		/*
		Graphics2D g2 = (Graphics2D)g;
		if (basicStroke == null){
			basicStroke =  (new BasicStroke(width, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND ));
		}
		g2.setStroke(basicStroke);
		*/
		
		g.setColor(color);
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if(p0!=null && p1!=null){
			int x0 = x + p0.x;
			int y0 = y + p0.y;
			int x1 = x + p1.x;
			int y1 = y + p1.y;
			g.drawLine(x0, y0, x1, y1);
			}
		}
		if (points.size() == 1){
			Point p0 = (Point)points.elementAt(0);
			int x0 = x + p0.x;
			int y0 = y + p0.y;

			g.drawLine(x0, y0, x0, y0);
		}
	}
	public void scaled_paint(Graphics g, int x, int y, double scale){
		if (color == null)
			return;

		g.setColor(color);
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if(p0!=null && p1!=null){
			int x0 = x + (int)((p0.x-x)*scale);
			int y0 = y + (int)((p0.y-y)*scale);
			int x1 = x + (int)((p1.x-x)*scale);
			int y1 = y + (int)((p1.y-y)*scale);
			g.drawLine(x0, y0, x1, y1);
			}
		}
		if (points.size() == 1){
			Point p0 = (Point)points.elementAt(0);
			int x0 = x + (int)((p0.x-x)*scale);
			int y0 = y + (int)((p0.y-y)*scale);

			g.drawLine(x0, y0, x0, y0);
		}
	}
	public void paint(Graphics g, double sx, double sy, double s){
		if (color == null)
			return;
		

		g.setColor(color);
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if(p0!=null && p1!=null){
			int x0 = (int)(sx + (p0.x)*s);
			int y0 = (int)(sy + (p0.y)*s);
			int x1 = (int)(sx + (p1.x)*s);
			int y1 = (int)(sy + (p1.y)*s);
			g.drawLine(x0, y0, x1, y1);
			}
		}
		if (points.size() == 1){
			Point p0 = (Point)points.elementAt(0);
			int x0 = (int)(sx + (p0.x)*s);
			int y0 = (int)(sy + (p0.y)*s);
			g.drawLine(x0, y0, x0, y0);
		}
			
	}
	
	
	public int length(){
		int length = 0;
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if (p0!=null && p1!=null);
			length += (int) Vector2D.distance(p0,p1);
		}
		return length;
	}
	public int direct_length(){
		if (points.size()<2)
			return 0;
		Point p0 = (Point)points.elementAt(0);
		Point p1 = (Point)points.elementAt(points.size()-1);
		return (int) Vector2D.distance(p0,p1);
	}
	public boolean straight(){
		return (1.0*length()/direct_length()<1.2);
	}
		
	// self
	public void move_relative(int dx, int dy){
		for(int i=0; i< points.size(); i++){
			Point p = (Point)points.elementAt(i);
			if (p!=null){
				p.x += dx;
				p.y += dy;
			}
		}
		bbox.x += dx;
		bbox.y += dy;
	}
	// new 
	public Stroke shift_stroke_relative(int dx, int dy){
		Stroke s = new Stroke();
		s.width = width;
		for(int i=0; i< points.size(); i++){
			Point p = (Point)points.elementAt(i);
			if (p!=null)
				s.addPoint(p.x+dx, p.y+dy);
			else
				s.addPoint(null);
		}
		s.bbox = new Rectangle(bbox.x+dx, bbox.y+dy, bbox.width, bbox.height);
		return s;
	}
	

	
	
	
	
	
	
	
	
	
	public void update_bbox(){
		if (points.size() == 0){
			bbox = new Rectangle(0,0,0,0);
			return;
		}
		
		int x0=-1, y0=-1, x1=-1, y1=-1;
		for(int i=0; i< points.size(); i++){
			Point point = (Point) points.elementAt(i);
			if (x0 == -1 || x0 > point.x)
				x0 = point.x;
			if (y0 == -1 || y0 > point.y)
				y0 = point.y;
			if (x1 == -1 || x1 < point.x)
				x1 = point.x;
			if (y1 == -1 || y1 < point.y)
				y1 = point.y;
		}
		
		
		bbox = new Rectangle(x0, y0, x1-x0, y1-y0);
	}
	
	
	
	
	
	
	
	
	
	
	

	public void paint_graph(Graphics g){
		if (color != Color.red)
			return;
		int y0 = bbox.y+bbox.height+ 20;
		int x0 = bbox.x;
		int h = 50;
		int w = 50;
		
		g.setColor(Color.yellow);
		g.drawRect(x0, y0, w, h);
		
		g.setColor(Color.red);
		double total_length = length();
		Vector2D prev_vec = new Vector2D(1,0);
		double angle = 0;
		double x=x0;
		double prev_x = x0;
		for(int i=0; i< points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			Vector2D vec = new Vector2D(p0, p1);
			angle += prev_vec.get_angle180(vec);
			x += (w*Vector2D.distance(p0, p1)/total_length);
			double m = angle/360;
			m = m-Math.floor(m);
			int y = (int)( y0+h*m);//angle/360);
			g.drawLine((int)prev_x, y, (int)x, y);
			
			prev_x = x;
			prev_vec = vec;
		}
	
	
	}
	
	
	public boolean closedBy(Stroke loop){
		for(int i=0; i<points.size(); i++){
			Point p = (Point)points.elementAt(i);
			if (!loop.enclose(p))
				return false;
		}
		return true;
	}

	public boolean enclose(Point p){
		Point q0 = (Point)points.elementAt(points.size()-1);
		double angle = 0;
		for(int i=0; i<points.size(); i++){
			Point q1 = (Point)points.elementAt(i);
			Vector2D v0 = new Vector2D(p, q0);
			Vector2D v1 = new Vector2D(p, q1);
			angle += v0.get_angle180(v1);
			q0 = q1;
		}
		return (angle <-180 || angle > 180);
	}
	
	public boolean intersects(Stroke stroke){
		for(int i=0; i<points.size()-1; i++){
			Point p0 = (Point)points.elementAt(i);
			Point p1 = (Point)points.elementAt(i+1);
			if (stroke.intersects(p0,p1))
				return true;
		}
		return false;
	}
	public boolean intersects(Point p0, Point p1){
		for(int i=0; i<points.size()-1; i++){
			Point q0 = (Point)points.elementAt(i);
			Point q1 = (Point)points.elementAt(i+1);
			if (intersects(p0.x, p0.y, p1.x, p1.y, q0.x, q0.y, q1.x, q1.y))  
				return true;
		}
		return false;
		
	}
    // ƂČ_邩B
    public boolean intersects(
		double x1,  double y1,  double x2,  double y2,
		double xx1, double yy1, double xx2, double yy2){

	  // a * x + b * y + c = 0
	  double a0, b0, c0,
        	 a1, b1, c1; 
	  a0 = y1 - y2; 
	  b0 = x2 - x1;
	  c0 = y2 * x1 - x2 * y1;
	  a1 = yy1 - yy2; 
	  b1 = xx2 - xx1;
	  c1 = yy2 * xx1 - xx2 * yy1;

	  if (((a0* xx1+b0* yy1+c0)*(a0* xx2+b0* yy2+c0)<= 0) &&
	      ((a1*  x1+b1*  y1+c1)*(a1*  x2+b1*  y2+c1)<= 0))
		return true;
	  else
		return false;
   }

	
	
	
	
	public void save(BufferedWriter out) throws IOException{
		out.write("stroke");
		out.newLine();
		
		out.write(save());
		out.newLine();
	}
	public String save(){
		String line = "";
		for(int i=0; i<points.size(); i++){
			Point p = (Point)points.elementAt(i);
			line = line + p.x+" "+p.y+" ";
		}
		return line;
	}
	
	
	
	
	public void load(BufferedReader in) throws IOException{
		String line = in.readLine();
		load(line);
	}
	public void load(String line) {
		StringTokenizer st = new StringTokenizer(line);
		String token;
		while(st. hasMoreTokens()){
			token = st.nextToken();		
			int x = (new Integer(token)).intValue();
			if (!st.hasMoreTokens())
				break;
			token = st.nextToken();
			int y = (new Integer(token)).intValue();
			points.addElement(new Point(x,y));
		}
		update_bbox();
	}

	public Polygon get_polygon(){
		int n = points.size();
		int[] x = new int[n];
		int[] y = new int[n];
		for(int i=0; i<points.size(); i++){
			Point p = (Point) points.elementAt(i);
			x[i] = p.x;
			y[i] = p.y;
		}
		return new Polygon(x,y,n);	
	}
	
	
	
	
	public void setColor(Color color){
		this.color = color;
	}
	public Point getPoint(int n){
		return (Point) points.elementAt(n);
	}
	public int getNumPoints(){
		return points.size();
	}
	public void removeAllPoints(){
		points = new Vector();
	}
	
	
	
	public double distance(Point p){
		return StrokeAnalyzer.distance(this, p);
	}
}
